#!/usr/bin/python
#
# calibcant - tools for thermally calibrating AFM cantilevers
#
# Copyright (C) 2008-2010 W. Trevor King <wking@drexel.edu>
#
# This file is part of CalibCant.
#
# CalibCant is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# CalibCant is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with CalibCant.  If not, see
# <http://www.gnu.org/licenses/>.

"""
Separate the more general T_analyze() from the other T_*()
functions in calibcant.  Also provide a command line interface
for analyzing data acquired through other workflows.

The relevant physical quantities are :
 T Temperature at which thermal vibration measurements were aquired
"""

import numpy

import data_logger
from splittable_kwargs import splittableKwargsFunction, \
    make_splittable_kwargs_function

from . import common
from . import config


def C_to_K(celsius) :
    "Convert Celsius -> Kelvin."
    return celsius + 273.15

def K_to_K(kelvin) :
    "Convert Kelvin -> Kelvin."
    return kelvin

@splittableKwargsFunction()
def T_analyze(T, convert_to_K=C_to_K) :
    """
    Not much to do here, just convert to Kelvin.
    Uses C_to_K (defined above) by default.
    """
    try : # if T is an array, convert element by element
        for i in range(len(T)) :
            T[i] = convert_to_K(T[i])
    except TypeError : # otherwise, make an array from a single T
        T = [convert_to_K(T)]
    return T

@splittableKwargsFunction()
def T_save(T, log_dir=None) :
    """
    Save either a single T (if you are crazy :p),
    or an array of them (more reasonable)
    """
    T = numpy.array(T, dtype=numpy.float)
    if log_dir != None :
        log = data_logger.data_log(log_dir, noclobber_logsubdir=False,
                                   log_name="T_float")
        log.write_binary(T.tostring())
    if config.LOG_DATA != None :
        log = data_logger.data_log(config.LOG_DIR, noclobber_logsubdir=False,
                                   log_name="T_float")
        log.write_binary(T.tostring())
        
def T_load(datafile=None) :
    """
    Load the saved T array (possibly of length 1), and return it.  If
    datafile == None, return an array of one config.DEFAULT_TEMP
    instead.
    """
    if datafile == None :
        return numpy.array([config.DEFAULT_TEMP], dtype=numpy.float)
    else :
        dl = data_logger.data_load()
        return dl.read_binary(datafile)

@splittableKwargsFunction()
def T_plot(T, plotVerbose=False) :
    """
    Umm, just print the temperature?
    """
    if plotVerbose or config.PYLAB_VERBOSE or config.TEXT_VERBOSE :
        print "Temperature ", T

@splittableKwargsFunction((T_analyze, 'T', 'convert_to_K'),
                          (T_plot, 'T'))
def T_load_analyze_tweaked(tweak_file, convert_to_K=C_to_K, textVerboseFile=None, **kwargs) :
    "Load all the T array files from a tweak file and return a single array"
    T_analyze_kwargs,T_plot_kwargs = \
        T_load_analyze_tweaked._splitargs(T_load_analyze_tweaked, kwargs)
    Ts = []
    for line in file(tweak_file, 'r') :
        parsed = line.split()
        path = parsed[0].strip()
        if path[0] == '#': # a comment
            continue
        if textVerboseFile != None :
            print >> textVerboseFile, "Reading data from %s" % (path)
        # read the data
        data = T_load(path)
        Ts.extend(data)
    T_analyze(Ts, convert_to_K=convert_to_K)
    return numpy.array(Ts, dtype=numpy.float)

# commandline interface functions
import scipy.io, sys

def read_data(ifile):
    "ifile can be a filename string or open (seekable) file object"
    if ifile == None :  ifile = sys.stdin
    unlabeled_data=scipy.io.read_array(ifile)
    return unlabeled_data

if __name__ == '__main__' :
    # command line interface
    from optparse import OptionParser
    
    usage_string = ('%prog <input-file>\n'
                    '2008, W. Trevor King.\n'
                    '\n'
                    'There are two operation modes, one to analyze a single T (temperature) file,\n'
                    'and one to analyze tweak files.\n'
                    '\n'
                    'Single file mode (the default) :\n'
                    'Reads in single column ASCII file of temperatures and... prints them back out.\n'
                    'No need to do this, but the option is available for consistency with the other\n'
                    'calibcant modules.\n'
                    '\n'
                    'Tweak file mode:\n'
                    'Runs the same analysis as in single file mode for each T file in\n'
                    'a tweak file.  Each line in the tweak file specifies a single T file.\n'
                    'Blank lines and those beginning with a pound sign (#) are ignored.\n'
                    'A T file contains a sequence of 32 bit floats representing temperature in K.\n'
                    )
    parser = OptionParser(usage=usage_string, version='%prog 0.1')
    parser.add_option('-C', '--celsius', dest='celsius',
                      help='Use Celsius input temperatures instead of Kelvin (default %default)\n',
                      action='store_true', default=False)
    parser.add_option('-o', '--output-file', dest='ofilename',
                      help='write output to FILE (default stdout)',
                      type='string', metavar='FILE')
    parser.add_option('-c', '--comma-out', dest='comma_out', action='store_true',
                      help='Output comma-seperated values (default %default)',
                      default=False)
    parser.add_option('-t', '--tweak-mode', dest='tweakmode', action='store_true',
                      help='Run in tweak-file mode',
                      default=False)
    parser.add_option('-v', '--verbose', dest='verbose', action='store_true',
                      help='Print lots of debugging information',
                      default=False)

    options,args = parser.parse_args()
    parser.destroy()
    assert len(args) >= 1, "Need an input file"
        
    ifilename = args[0]

    if options.ofilename != None :
        ofile = file(options.ofilename, 'w')
    else :
        ofile = sys.stdout
    if options.verbose == True :
        vfile = sys.stderr
    else :
        vfile = None
    config.TEXT_VERBOSE = options.verbose
    config.PYLAB_VERBOSE = False
    config.GNUPLOT_VERBOSE = False
    if options.celsius :
        convert_to_K = C_to_K
    else :
        convert_to_K = K_to_K
    
    if options.tweakmode == False :
        data = read_data(ifilename)
        Ts = T_analyze(data, convert_to_K)
    else : # tweak file mode
        Ts = T_load_analyze_tweaked(ifilename, convert_to_K, textVerboseFile=vfile)

    if options.comma_out :
        sep = ','
    else :
        sep = '\n'
    common.write_array(ofile, Ts, sep)
    
    if options.ofilename != None :
        ofile.close()
