#!/usr/bin/python
#
# calibcant - tools for thermally calibrating AFM cantilevers
#
# Copyright (C) 2008-2010 W. Trevor King <wking@drexel.edu>
#
# This file is part of CalibCant.
#
# CalibCant is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# CalibCant is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with CalibCant.  If not, see
# <http://www.gnu.org/licenses/>.

"""
Separate the more general calib_analyze() from the other calib_*()
functions in calibcant.  Also provide a command line interface
for analyzing data acquired through other workflows.

The relevent physical quantities are : 
 Vzp_out  Output z-piezo voltage (what we generate)
 Vzp      Applied z-piezo voltage (after external ZPGAIN)
 Zp       The z-piezo position
 Zcant    The cantilever vertical deflection
 Vphoto   The photodiode vertical deflection voltage (what we measure)
 Fcant    The force on the cantilever
 T        The temperature of the cantilever and surrounding solution
          (another thing we measure)
 k_b      Boltzmann's constant

Which are related by the parameters :
 zpGain           Vzp_out / Vzp
 zpSensitivity    Zp / Vzp
 photoSensitivity Vphoto / Zcant
 k_cant           Fcant / Zcant
"""

import numpy

import data_logger
from splittable_kwargs import splittableKwargsFunction, \
    make_splittable_kwargs_function

from . import common
from . import config
from . import T_analyze


kb = 1.3806504e-23 # Boltzmann's constant in J/K

#@splittableKwargsFunction((calib_plot, 'bumps', 'Ts', 'vibs'))
# Some of the child functions aren't yet defined, so postpone
# make-splittable until later in the module.
def calib_analyze(bumps, Ts, vibs, **kwargs) :
    """
    Analyze data from get_calibration_data()
    return (k, k_s, !!!a2_r, T_r, one_o_Vp2_r)
    Inputs (all are arrays of recorded data) :
     bumps measured (V_photodiode / nm_tip) proportionality constant
     Ts    measured temperature (K)
     vibs  measured V_photodiode variance in free solution (V**2)
    Outputs :
     k    cantilever spring constant (in N/m, or equivalently nN/nm)
     k_s  standard deviation in our estimate of k
     !!!a2_r relative error in a**2
     !!!T_r  relative error in T
     !!!one_o_Vp2_r relative error in 1/Vphotodiode_variance
    Notes :
     We're assuming vib is mostly from thermal cantilever vibrations
    (and then only from vibrations in the single vertical degree of freedom),
    and not from other noise sources.
     The various relative errors are returned to help you gauge the
    largest source of random error in your measurement of k.
    If one of them is small, don't bother repeating that measurment too often.
    If one is large, try repeating that measurement more.
    Remember that you need enough samples to have a valid error estimate in
    the first place, and that none of this addresses any systematic errors.
    """
    calib_plot_kwargs, = calib_analyze._splitargs(calib_analyze, kwargs)
    photoSensitivity2 = bumps**2
    one_o_Vphoto2 = 1/vibs

    photoSensitivity2_m = photoSensitivity2.mean()
    T_m = Ts.mean()
    one_o_Vphoto2_m = one_o_Vphoto2.mean()
    # Vphoto / photoSensitivity = x
    # k = kb T / <x**2> = kb T photoSensitiviy**2 * (1e9nm/m)**2 / 
    #                                                       <Vphoto_std**2>
    #
    # units,  photoSensitivity =  Vphoto/(Zcant in nm),
    # so Vphoto/photoSensitivity = Zcant in nm
    # so k = J/K * K / nm^2 * (1e9nm/m)**2 = N/m
    k  = kb * T_m * photoSensitivity2_m * one_o_Vphoto2_m * 1e18

    # propogation of errors !!!
    # first, get standard deviations
    photoSensitivity2_s = photoSensitivity2.std()
    T_s = Ts.std()
    one_o_Vphoto2_s = one_o_Vphoto2.std()
    # !!!!now, get relative errors
    photoSensitivity2_r = photoSensitivity2_s / photoSensitivity2_m
    T_r = T_s / T_m
    one_o_Vphoto2_r = one_o_Vphoto2_s / one_o_Vphoto2_m

    k_s = k*(photoSensitivity2_r**2 + T_r**2 + one_o_Vphoto2_r**2)**0.5

    calib_plot(bumps, Ts, vibs, **calib_plot_kwargs)
    
    return (k, k_s,
            photoSensitivity2_m, photoSensitivity2_s,
            T_m, T_s, one_o_Vphoto2_m, one_o_Vphoto2_s)

@splittableKwargsFunction()
def string_errors(k_m, k_s,
                  photoSensitivity2_m, photoSensitivity2_s,
                  T_m, T_s,
                  one_o_Vphoto2_m, one_o_Vphoto2_s) :
    k_r = k_s / k_m
    photoSensitivity2_r = photoSensitivity2_s / photoSensitivity2_m
    T_r = T_s / T_m
    one_o_Vphoto2_r = one_o_Vphoto2_s / one_o_Vphoto2_m
    string  = "Variable (units)              : mean +/- std. dev. (relative error)\n"
    string += "Cantilever k (N/m)            : %g +/- %g (%g)\n" \
              % (k_m, k_s, k_r)
    string += "photoSensitivity**2 (V/nm)**2 : %g +/- %g (%g)\n" \
              % (photoSensitivity2_m, photoSensitivity2_s, photoSensitivity2_r)
    string += "T (K)                         : %g +/- %g (%g)\n" \
              % (T_m, T_s, T_r)
    string += "1/Vphoto**2 (1/V)**2          : %g +/- %g (%g)" \
              % (one_o_Vphoto2_m, one_o_Vphoto2_s, one_o_Vphoto2_r)
    return string

@splittableKwargsFunction()
def calib_save(bumps, Ts, vibs, log_dir=None) :
    """
    Save a dictonary with the bump, T, and vib data.
    """
    if log_dir != None :
        data = {'bump':bumps, 'T':Ts, 'vib':vibs}
        log = data_logger.data_log(log_dir, noclobber_logsubdir=False,
                                   log_name="calib")
        log.write_dict_of_arrays(data)

def calib_load(datafile) :
    "Load the dictionary data, using data_logger.date_load()"
    dl = data_logger.data_load()
    data = dl.read_dict_of_arrays(path)
    return (data['bump'], data['T'], data['vib'])

def calib_save_analysis(k, k_s,
                        photoSensitivity2_m, photoSensitivity2_s,
                        T_m, T_s, one_o_Vphoto2_m, one_o_Vphoto2_s,
                        log_dir=None) :
    if log_dir != None :
        log = data_logger.data_log(log_dir, noclobber_logsubdir=False,
                                   log_name="calib_analysis_text")
        log.write_binary(string_errors(k, k_s,
                                       photoSensitivity2_m, photoSensitivity2_s,
                                       T_m, T_s,
                                       one_o_Vphoto2_m, one_o_Vphoto2_s)
                         +'\n')

@splittableKwargsFunction()
def calib_plot(bumps, Ts, vibs, plotVerbose=False) :
    if plotVerbose or config.PYLAB_VERBOSE :
        common._import_pylab()
        common._pylab.figure(config.BASE_FIGNUM+4)
        common._pylab.subplot(311)
        common._pylab.plot(bumps, 'g.-')
        common._pylab.title('Photodiode sensitivity (V/nm)')
        common._pylab.subplot(312)
        common._pylab.plot(Ts, 'r.-')
        common._pylab.title('Temperature (K)')
        common._pylab.subplot(313)
        common._pylab.plot(vibs, 'b.-')
        common._pylab.title('Thermal deflection variance (Volts**2)')
        common._flush_plot()

make_splittable_kwargs_function(calib_analyze,
                                (calib_plot, 'bumps', 'Ts', 'vibs'))

@splittableKwargsFunction((calib_analyze, 'bumps', 'Ts', 'vibs'))
def calib_load_analyze_tweaks(bump_tweaks, vib_tweaks, T_tweaks=None) :
    raise NotImplementedError
    a = read_tweaked_bumps(bump_tweaks)
    vib = V_photo_variance_from_file(vib_tweaks)
    if T_tweaks == None:
        pass
    return analyze_calibration_data(a, T, vib, log_dir=log_dir)

# commandline interface functions
import scipy.io, sys

def array_from_string(string):
    ret = []
    for num in string.split(',') :
        ret.append(float(num))
    assert len(ret) > 0
    return numpy.array(ret)

def read_data(ifile):
    "ifile can be a filename string or open (seekable) file object"
    unlabeled_data=scipy.io.read_array(file)
    return unlabeled_data

def get_array(string, filename, name) :
    "get an array from supplied command line options"
    if string != None :
        array = array_from_string(string)
    elif filename != None :
        array = read_data(filename)
    else :
        raise Exception, "no %s information given" % (name)
    return array

if __name__ == '__main__' :
    # command line interface
    from optparse import OptionParser
    
    usage_string = ('%prog <bumps> <temps> <vibs>\n'
                    '2008, W. Trevor King.\n'
                    '\n'
                    'Takes arrays of Vphotodiode sensitivity (V/nm), Temperature (K), \n'
                    'and Vibration variance (V**2) as comma seperated lists.\n'
                    'Returns the cantilever spring constant (pN/nm).\n'
                    'for example:\n'
                    ' $ %prog -b 0.02,0.03,0.025 -t 298.2,300.1 -v 6e-9,5.5e-9\n'
                    )
    parser = OptionParser(usage=usage_string, version='%prog '+common.VERSION)
    parser.add_option('-b', '--bump-string', dest='bump_string',
                      help='comma seperated photodiode sensitivities (V/nm)',
                      type='string', metavar='BUMPS')
    parser.add_option('-t', '--temp-string', dest='temp_string',
                      help='comma seperated temperatures (K)',
                      type='string', metavar='TEMPS')
    parser.add_option('-v', '--vib-string', dest='vib_string',
                      help='comma seperated vibration variances (V**2)',
                      type='string', metavar='VIBS')
    parser.add_option('-B', '--bump-file', dest='bump_file',
                      help='comma seperated photodiode sensitivities (V/nm)',
                      type='string', metavar='BUMPFILE')
    parser.add_option('-T', '--temp-file', dest='temp_file',
                      help='comma seperated temperatures (K)',
                      type='string', metavar='TEMPFILE')
    parser.add_option('-V', '--vib-file', dest='vib_file',
                      help='comma seperated vibration variances (V**2)',
                      type='string', metavar='VIBFILE')
    parser.add_option('-C', '--celsius', dest='celsius',
                      help='Use Celsius input temperatures instead of Kelvin (defaul %default)\n',
                      action='store_true', default=False)
    parser.add_option('-o', '--output-file', dest='ofilename',
                      help='write output to FILE (default stdout)',
                      type='string', metavar='FILE')
    parser.add_option('-p', '--plot-inputs', dest='plot',
                      help='plot the input arrays to check their distribution',
                      action='store_true', default=False)
    parser.add_option('', '--verbose', dest='verbose', action='store_true',
                      help='print lots of debugging information',
                      default=False)

    options,args = parser.parse_args()
    parser.destroy()
    
    config.TEXT_VERBOSE = options.verbose
    config.PYLAB_INTERACTIVE = False
    config.PYLAB_VERBOSE = options.plot
    
    bumps = get_array(options.bump_string, options.bump_file, "bump")
    temps = get_array(options.temp_string, options.temp_file, "temp")
    vibs = get_array(options.vib_string, options.vib_file, "vib")

    #if options.plot :
    #    calib_plot(bumps, temps, vibs)

    if options.celsius :
        for i in range(len(temps)) :
            temps[i] = T_analyze.C_to_K(temps[i])

    km,ks,ps2m,ps2s,Tm,Ts,ooVp2m,ooVp2s = \
        calib_analyze(bumps, temps, vibs, plotVerbose=options.plot)

    if options.verbose :
        print >> sys.stderr, \
            string_errors(km,ks,ps2m,ps2s,Tm,Ts,ooVp2m,ooVp2s)

    if options.ofilename != None :
        print >> file(options.ofilename, 'w'), km
    else :
        print km
